
/* Chrysalide - Outil d'analyse de fichiers binaires
 * operand.c - gestion générique des opérandes
 *
 * Copyright (C) 2008-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "operand.h"


#include <assert.h>
#include <malloc.h>
#include <string.h>


#include "operand-int.h"
#include "../common/sort.h"



/* Initialise la classe générique des opérandes. */
static void g_arch_operand_class_init(GArchOperandClass *);

/* Initialise une instance d'opérande d'architecture. */
static void g_arch_operand_init(GArchOperand *);

/* Procède à l'initialisation de l'interface de partage. */
static void g_arch_operand_interface_init(GSharedInstanceInterface *);

/* Supprime toutes les références externes. */
static void g_arch_operand_dispose(GArchOperand *);

/* Procède à la libération totale de la mémoire. */
static void g_arch_operand_finalize(GArchOperand *);

/* Initialise un nouvel objet partagé avec des informations. */
static bool g_arch_operand_do_init(GArchOperand *, const void *);

/* Fournit la valeur du compteur de partage. */
static unsigned int g_arch_operand_get_references(const GArchOperand *);

/* Incrémente le compteur de partage. */
static void g_arch_operand_inc_references(GArchOperand *);

/* Décrémente le compteur de partage. */
static void g_arch_operand_dec_references(GArchOperand *);

/* Compare de façon accélérée un opérande avec un autre. */
static int g_arch_operand_quickly_compare(const GArchOperand **, const GArchOperand **);


/* Indique le type défini pour un opérande d'architecture. */
G_DEFINE_TYPE_WITH_CODE(GArchOperand, g_arch_operand, G_TYPE_OBJECT,
                        G_IMPLEMENT_INTERFACE(G_TYPE_SHARED_INSTANCE, g_arch_operand_interface_init));



/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe générique des opérandes.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arch_operand_class_init(GArchOperandClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_arch_operand_dispose;
    object->finalize = (GObjectFinalizeFunc)g_arch_operand_finalize;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance à initialiser.                            *
*                                                                             *
*  Description : Initialise une instance d'opérande d'architecture.           *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arch_operand_init(GArchOperand *operand)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : iface = interface GLib à initialiser.                        *
*                                                                             *
*  Description : Procède à l'initialisation de l'interface de partage.        *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arch_operand_interface_init(GSharedInstanceInterface *iface)
{
    iface->init = (init_shared_fc)g_arch_operand_do_init;

    iface->get_ref = (get_shared_ref_fc)g_arch_operand_get_references;
    iface->inc_ref = (inc_shared_ref_fc)g_arch_operand_inc_references;
    iface->dec_ref = (dec_shared_ref_fc)g_arch_operand_dec_references;

    iface->qck_cmp = (qck_compare_shared_fc)g_arch_operand_quickly_compare;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arch_operand_dispose(GArchOperand *operand)
{
    G_OBJECT_CLASS(g_arch_operand_parent_class)->dispose(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arch_operand_finalize(GArchOperand *operand)
{
	g_arch_operand_set_alt_text(operand, NULL, RTT_COUNT);

    G_OBJECT_CLASS(g_arch_operand_parent_class)->finalize(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = objet partagé à initialiser.                       *
*                info    = information à utiliser pour la mise en place.      *
*                                                                             *
*  Description : Initialise un nouvel objet partagé avec des informations.    *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_arch_operand_do_init(GArchOperand *operand, const void *info)
{
    bool result;                            /* Bilan à retourner           */

    result = G_ARCH_OPERAND_GET_CLASS(operand)->init(G_SHARED_INSTANCE(operand), info);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = objet partagé à consulter.                         *
*                                                                             *
*  Description : Fournit la valeur du compteur de partage.                    *
*                                                                             *
*  Retour      : Nombre de partages courant.                                  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static unsigned int g_arch_operand_get_references(const GArchOperand *operand)
{
    return operand->shared_count;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = objet partagé à modifier.                          *
*                                                                             *
*  Description : Incrémente le compteur de partage.                           *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arch_operand_inc_references(GArchOperand *operand)
{
    operand->shared_count++;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = objet partagé à modifier.                          *
*                                                                             *
*  Description : Décrémente le compteur de partage.                           *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arch_operand_dec_references(GArchOperand *operand)
{
    operand->shared_count--;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : a = premier opérande à consulter.                            *
*                b = second opérande à consulter.                             *
*                                                                             *
*  Description : Compare de façon accélérée un opérande avec un autre.        *
*                                                                             *
*  Retour      : Bilan de la comparaison.                                     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int g_arch_operand_quickly_compare(const GArchOperand **a, const GArchOperand **b)
{
    int result;                             /* Bilan à faire remonter      */

    result = G_ARCH_OPERAND_GET_CLASS(*a)->compare(a, b);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : a = premier opérande à consulter.                            *
*                b = second opérande à consulter.                             *
*                                                                             *
*  Description : Compare un opérande avec un autre.                           *
*                                                                             *
*  Retour      : Bilan de la comparaison.                                     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

int g_arch_operand_compare(const GArchOperand * const *a, const GArchOperand * const *b)
{
    int result;                             /* Bilan à faire remonter      */
    GType type_a;                           /* Type de l'object A          */
    GType type_b;                           /* Type de l'object B          */

    type_a = G_OBJECT_TYPE(G_OBJECT(*a));
    type_b = G_OBJECT_TYPE(G_OBJECT(*b));

    assert(sizeof(GType) <= sizeof(unsigned long));

    result = sort_unsigned_long(type_a, type_b);

    if (result == 0)
        result = G_ARCH_OPERAND_GET_CLASS(*a)->compare(a, b);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à traiter.                                *
*                text    = représentation lisible alternative.                *
*                tag     = style d'impression pour le remplacement.           *
*                                                                             *
*  Description : Définit une autre représentation textuelle pour l'opérande.  *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_arch_operand_set_alt_text(GArchOperand *operand, const char *text, RenderingTagType tag)
{
    size_t alt_len;                         /* Taille du texte alternatif  */

    if (operand->alt_info != NULL)
        free(operand->alt_info);

    if (text == NULL)
        operand->alt_info = NULL;

    else
    {
        alt_len = strlen(text);

        operand->alt_info = (alt_rendering *)malloc(sizeof(RenderingTagType) + alt_len + 1);

        operand->alt_info->tag = tag;
        strcpy(operand->alt_info->text, text);

    }

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à traiter.                                *
*                line    = ligne tampon où imprimer l'opérande donné.         *
*                syntax  = type de représentation demandée.                   *
*                                                                             *
*  Description : Traduit un opérande en version humainement lisible.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_arch_operand_print(const GArchOperand *operand, GBufferLine *line, AsmSyntax syntax)
{
    size_t alt_len;                         /* Taille du texte alternatif  */

    if (operand->alt_info != NULL)
    {
        alt_len = strlen(operand->alt_info->text);

        g_buffer_line_append_text(line, BLC_ASSEMBLY,
                                  operand->alt_info->text,
                                  alt_len,
                                  operand->alt_info->tag,
                                  NULL);

    }
    else
        G_ARCH_OPERAND_GET_CLASS(operand)->print(operand, line, syntax);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à consulter.                              *
*                binary  = informations relatives au binaire chargé.          *
*                                                                             *
*  Description : Construit un petit résumé concis de l'opérande.              *
*                                                                             *
*  Retour      : Chaîne de caractères à libérer après usage ou NULL.          *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

char *g_arch_operand_build_tooltip(const GArchOperand *operand, const GLoadedBinary *binary)
{
    char *result;                           /* Description à retourner     */
    GArchOperandClass *class;               /* Classe associée à l'objet   */

    class = G_ARCH_OPERAND_GET_CLASS(operand);

    if (class->build_tooltip != NULL)
        result = class->build_tooltip(operand, binary);
    else
        result = NULL;

    return result;

}
